@@ -2,7 +2,7 @@
  * Copyright (c) 1990 - 1994, Julianne Frances Haugh
  * Copyright (c) 1996 - 2001, Marek Michałkiewicz
  * Copyright (c) 2001 - 2006, Tomasz Kłoczko
- * Copyright (c) 2007 - 2008, Nicolas François
+ * Copyright (c) 2007 - 2011, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -32,9 +32,10 @@
 
 #include <config.h>
 
-#ident "$Id: commonio.c 2645 2009-04-11 16:08:00Z nekral-guest $"
+#ident "$Id: commonio.c 3655 2011-12-09 22:13:02Z nekral-guest $"
 
 #include "defines.h"
+#include <assert.h>
 #include <sys/stat.h>
 #include <stdlib.h>
 #include <limits.h>
@@ -44,27 +45,34 @@
 #include <stdio.h>
 #include <signal.h>
 #include "nscd.h"
-#ifdef WITH_SELINUX
-#include <selinux/selinux.h>
-static security_context_t old_context = NULL;
-#endif
+#ifdef WITH_TCB
+#include <tcb.h>
+#endif				/* WITH_TCB */
+#include "prototypes.h"
 #include "commonio.h"
 
 /* local function prototypes */
 static int lrename (const char *, const char *);
 static int check_link_count (const char *file);
-static int do_lock_file (const char *file, const char *lock);
-static FILE *fopen_set_perms (const char *, const char *, const struct stat *);
+static int do_lock_file (const char *file, const char *lock, bool log);
+static /*@null@*/ /*@dependent@*/FILE *fopen_set_perms (
+	const char *name,
+	const char *mode,
+	const struct stat *sb);
 static int create_backup (const char *, FILE *);
 static void free_linked_list (struct commonio_db *);
-static void add_one_entry (struct commonio_db *, struct commonio_entry *);
+static void add_one_entry (
+	struct commonio_db *db,
+	/*@owned@*/struct commonio_entry *p);
 static bool name_is_nis (const char *name);
 static int write_all (const struct commonio_db *);
-static struct commonio_entry *find_entry_by_name (struct commonio_db *,
-						  const char *);
-static struct commonio_entry *next_entry_by_name (struct commonio_db *,
-                                                  struct commonio_entry *pos,
-                                                  const char *);
+static /*@dependent@*/ /*@null@*/struct commonio_entry *find_entry_by_name (
+	struct commonio_db *,
+	const char *);
+static /*@dependent@*/ /*@null@*/struct commonio_entry *next_entry_by_name (
+	struct commonio_db *,
+	/*@null@*/struct commonio_entry *pos,
+	const char *);
 
 static int lock_count = 0;
 static bool nscd_need_reload = false;
@@ -75,21 +83,36 @@ static bool nscd_need_reload = false;
  */
 int lrename (const char *old, const char *new)
 {
-
-	char resolved_path[PATH_MAX];
 	int res;
+	char *r = NULL;
 
 #if defined(S_ISLNK)
+#ifndef __GLIBC__
+	char resolved_path[PATH_MAX];
+#endif				/* !__GLIBC__ */
 	struct stat sb;
 	if (lstat (new, &sb) == 0 && S_ISLNK (sb.st_mode)) {
-		if (realpath (new, resolved_path) == NULL) {
+#ifdef __GLIBC__ /* now a POSIX.1-2008 feature */
+		r = realpath (new, NULL);
+#else				/* !__GLIBC__ */
+		r = realpath (new, resolved_path);
+#endif				/* !__GLIBC__ */
+		if (NULL == r) {
 			perror ("realpath in lrename()");
 		} else {
-			new = resolved_path;
+			new = r;
 		}
 	}
-#endif
+#endif				/* S_ISLNK */
+
 	res = rename (old, new);
+
+#ifdef __GLIBC__
+	if (NULL != r) {
+		free (r);
+	}
+#endif				/* __GLIBC__ */
+
 	return res;
 }
 
@@ -109,7 +132,7 @@ static int check_link_count (const char *file)
 }
 
 
-static int do_lock_file (const char *file, const char *lock)
+static int do_lock_file (const char *file, const char *lock, bool log)
 {
 	int fd;
 	pid_t pid;
@@ -119,6 +142,11 @@ static int do_lock_file (const char *file, const char *lock)
 
 	fd = open (file, O_CREAT | O_EXCL | O_WRONLY, 0600);
 	if (-1 == fd) {
+		if (log) {
+			(void) fprintf (stderr,
+			                "%s: %s: %s\n",
+			                Prog, file, strerror (errno));
+		}
 		return 0;
 	}
 
@@ -126,6 +154,11 @@ static int do_lock_file (const char *file, const char *lock)
 	snprintf (buf, sizeof buf, "%lu", (unsigned long) pid);
 	len = (ssize_t) strlen (buf) + 1;
 	if (write (fd, buf, (size_t) len) != len) {
+		if (log) {
+			(void) fprintf (stderr,
+			                "%s: %s: %s\n",
+			                Prog, file, strerror (errno));
+		}
 		(void) close (fd);
 		unlink (file);
 		return 0;
@@ -134,12 +167,22 @@ static int do_lock_file (const char *file, const char *lock)
 
 	if (link (file, lock) == 0) {
 		retval = check_link_count (file);
+		if ((0==retval) && log) {
+			(void) fprintf (stderr,
+			                "%s: %s: lock file already used\n",
+			                Prog, file);
+		}
 		unlink (file);
 		return retval;
 	}
 
 	fd = open (lock, O_RDWR);
 	if (-1 == fd) {
+		if (log) {
+			(void) fprintf (stderr,
+			                "%s: %s: %s\n",
+			                Prog, lock, strerror (errno));
+		}
 		unlink (file);
 		errno = EINVAL;
 		return 0;
@@ -147,31 +190,60 @@ static int do_lock_file (const char *file, const char *lock)
 	len = read (fd, buf, sizeof (buf) - 1);
 	close (fd);
 	if (len <= 0) {
+		if (log) {
+			(void) fprintf (stderr,
+			                "%s: existing lock file %s without a PID\n",
+			                Prog, lock);
+		}
 		unlink (file);
 		errno = EINVAL;
 		return 0;
 	}
 	buf[len] = '\0';
-	/* FIXME: use a get_pid */
-	pid = strtol (buf, (char **) 0, 10);
-	if (0 == pid) {
+	if (get_pid (buf, &pid) == 0) {
+		if (log) {
+			(void) fprintf (stderr,
+			                "%s: existing lock file %s with an invalid PID '%s'\n",
+			                Prog, lock, buf);
+		}
 		unlink (file);
 		errno = EINVAL;
 		return 0;
 	}
 	if (kill (pid, 0) == 0) {
+		if (log) {
+			(void) fprintf (stderr,
+			                "%s: lock %s already used by PID %lu\n",
+			                Prog, lock, (unsigned long) pid);
+		}
 		unlink (file);
 		errno = EEXIST;
 		return 0;
 	}
 	if (unlink (lock) != 0) {
+		if (log) {
+			(void) fprintf (stderr,
+			                "%s: cannot get lock %s: %s\n",
+			                Prog, lock, strerror (errno));
+		}
 		unlink (file);
 		return 0;
 	}
 
 	retval = 0;
-	if ((link (file, lock) == 0) && (check_link_count (file) != 0)) {
-		retval = 1;
+	if (link (file, lock) == 0) {
+		retval = check_link_count (file);
+		if ((0==retval) && log) {
+			(void) fprintf (stderr,
+			                "%s: %s: lock file already used\n",
+			                Prog, file);
+		}
+	} else {
+		if (log) {
+			(void) fprintf (stderr,
+			                "%s: cannot get lock %s: %s\n",
+			                Prog, lock, strerror (errno));
+		}
 	}
 
 	unlink (file);
@@ -179,8 +251,10 @@ static int do_lock_file (const char *file, const char *lock)
 }
 
 
-static FILE *fopen_set_perms (const char *name, const char *mode,
-                              const struct stat *sb)
+static /*@null@*/ /*@dependent@*/FILE *fopen_set_perms (
+	const char *name,
+	const char *mode,
+	const struct stat *sb)
 {
 	FILE *fp;
 	mode_t mask;
@@ -196,26 +270,27 @@ static FILE *fopen_set_perms (const char *name, const char *mode,
 	if (fchown (fileno (fp), sb->st_uid, sb->st_gid) != 0) {
 		goto fail;
 	}
-#else
+#else				/* !HAVE_FCHOWN */
 	if (chown (name, sb->st_mode) != 0) {
 		goto fail;
 	}
-#endif
+#endif				/* !HAVE_FCHOWN */
 
 #ifdef HAVE_FCHMOD
 	if (fchmod (fileno (fp), sb->st_mode & 0664) != 0) {
 		goto fail;
 	}
-#else
+#else				/* !HAVE_FCHMOD */
 	if (chmod (name, sb->st_mode & 0664) != 0) {
 		goto fail;
 	}
-#endif
+#endif				/* !HAVE_FCHMOD */
 	return fp;
 
       fail:
-	fclose (fp);
-	unlink (name);
+	(void) fclose (fp);
+	/* fopen_set_perms is used for intermediate files */
+	(void) unlink (name);
 	return NULL;
 }
 
@@ -249,11 +324,13 @@ static int create_backup (const char *backup, FILE * fp)
 		}
 	}
 	if ((c != EOF) || (ferror (fp) != 0) || (fflush (bkfp) != 0)) {
-		fclose (bkfp);
+		(void) fclose (bkfp);
+		/* FIXME: unlink the backup file? */
 		return -1;
 	}
 	if (   (fsync (fileno (bkfp)) != 0)
 	    || (fclose (bkfp) != 0)) {
+		/* FIXME: unlink the backup file? */
 		return -1;
 	}
 
@@ -299,7 +376,7 @@ bool commonio_present (const struct commonio_db *db)
 }
 
 
-int commonio_lock_nowait (struct commonio_db *db)
+int commonio_lock_nowait (struct commonio_db *db, bool log)
 {
 	char file[1024];
 	char lock[1024];
@@ -311,7 +388,7 @@ int commonio_lock_nowait (struct commonio_db *db)
 	snprintf (file, sizeof file, "%s.%lu",
 	          db->filename, (unsigned long) getpid ());
 	snprintf (lock, sizeof lock, "%s.lock", db->filename);
-	if (do_lock_file (file, lock) != 0) {
+	if (do_lock_file (file, lock, log) != 0) {
 		db->locked = true;
 		lock_count++;
 		return 1;
@@ -335,17 +412,22 @@ int commonio_lock (struct commonio_db *db)
 	 */
 	if (0 == lock_count) {
 		if (lckpwdf () == -1) {
+			if (geteuid () != 0) {
+				(void) fprintf (stderr,
+				                "%s: Permission denied.\n",
+				                Prog);
+			}
 			return 0;	/* failure */
 		}
 	}
 
-	if (commonio_lock_nowait (db) != 0) {
+	if (commonio_lock_nowait (db, true) != 0) {
 		return 1;	/* success */
 	}
 
 	ulckpwdf ();
 	return 0;		/* failure */
-#else
+#else				/* !HAVE_LCKPWDF */
 	int i;
 
 	/*
@@ -362,16 +444,18 @@ int commonio_lock (struct commonio_db *db)
 		if (i > 0) {
 			sleep (LOCK_SLEEP);	/* delay between retries */
 		}
-		if (commonio_lock_nowait (db) != 0) {
+		if (commonio_lock_nowait (db, i==LOCK_TRIES-1) != 0) {
 			return 1;	/* success */
 		}
 		/* no unnecessary retries on "permission denied" errors */
 		if (geteuid () != 0) {
+			(void) fprintf (stderr, "%s: Permission denied.\n",
+			                Prog);
 			return 0;
 		}
 	}
 	return 0;		/* failure */
-#endif
+#endif				/* !HAVE_LCKPWDF */
 }
 
 static void dec_lock_count (void)
@@ -388,7 +472,7 @@ static void dec_lock_count (void)
 			}
 #ifdef HAVE_LCKPWDF
 			ulckpwdf ();
-#endif
+#endif				/* HAVE_LCKPWDF */
 		}
 	}
 }
@@ -422,10 +506,19 @@ int commonio_unlock (struct commonio_db *db)
 }
 
 
-static void add_one_entry (struct commonio_db *db, struct commonio_entry *p)
+/*
+ * Add an entry at the end.
+ *
+ * defines p->next, p->prev
+ * (unfortunately, owned special are not supported)
+ */
+static void add_one_entry (struct commonio_db *db,
+                           /*@owned@*/struct commonio_entry *p)
 {
+	/*@-mustfreeonly@*/
 	p->next = NULL;
 	p->prev = db->tail;
+	/*@=mustfreeonly@*/
 	if (NULL == db->head) {
 		db->head = p;
 	}
@@ -451,21 +544,27 @@ static bool name_is_nis (const char *name)
 #endif
 
 #if KEEP_NIS_AT_END
-static void add_one_entry_nis (struct commonio_db *, struct commonio_entry *);
+static void add_one_entry_nis (struct commonio_db *db,
+                               /*@owned@*/struct commonio_entry *newp);
 
 /*
  * Insert an entry between the regular entries, and the NIS entries.
+ *
+ * defines newp->next, newp->prev
+ * (unfortunately, owned special are not supported)
  */
-static void
-add_one_entry_nis (struct commonio_db *db, struct commonio_entry *newp)
+static void add_one_entry_nis (struct commonio_db *db,
+                               /*@owned@*/struct commonio_entry *newp)
 {
 	struct commonio_entry *p;
 
 	for (p = db->head; NULL != p; p = p->next) {
 		if (name_is_nis (p->eptr ? db->ops->getname (p->eptr)
 		                         : p->line)) {
+			/*@-mustfreeonly@*/
 			newp->next = p;
 			newp->prev = p->prev;
+			/*@=mustfreeonly@*/
 			if (NULL != p->prev) {
 				p->prev->next = newp;
 			} else {
@@ -489,9 +588,10 @@ int commonio_open (struct commonio_db *db, int mode)
 	char *cp;
 	char *line;
 	struct commonio_entry *p;
-	void *eptr;
+	void *eptr = NULL;
 	int flags = mode;
-	int buflen;
+	size_t buflen;
+	int fd;
 	int saved_errno;
 
 	mode &= ~O_CREAT;
@@ -508,10 +608,31 @@ int commonio_open (struct commonio_db *db, int mode)
 		return 0;
 	}
 
-	db->head = db->tail = db->cursor = NULL;
+	db->head = NULL;
+	db->tail = NULL;
+	db->cursor = NULL;
 	db->changed = false;
 
-	db->fp = fopen (db->filename, db->readonly ? "r" : "r+");
+	fd = open (db->filename,
+	             (db->readonly ? O_RDONLY : O_RDWR)
+	           | O_NOCTTY | O_NONBLOCK | O_NOFOLLOW);
+	saved_errno = errno;
+	db->fp = NULL;
+	if (fd >= 0) {
+#ifdef WITH_TCB
+		if (tcb_is_suspect (fd) != 0) {
+			(void) close (fd);
+			errno = EINVAL;
+			return 0;
+		}
+#endif				/* WITH_TCB */
+		db->fp = fdopen (fd, db->readonly ? "r" : "r+");
+		saved_errno = errno;
+		if (NULL == db->fp) {
+			(void) close (fd);
+		}
+	}
+	errno = saved_errno;
 
 	/*
 	 * If O_CREAT was specified and the file didn't exist, it will be
@@ -526,16 +647,7 @@ int commonio_open (struct commonio_db *db, int mode)
 	}
 
 	/* Do not inherit fd in spawned processes (e.g. nscd) */
-	fcntl(fileno(db->fp), F_SETFD, FD_CLOEXEC);
-
-#ifdef WITH_SELINUX
-	db->scontext = NULL;
-	if ((is_selinux_enabled () > 0) && (!db->readonly)) {
-		if (fgetfilecon (fileno (db->fp), &db->scontext) < 0) {
-			goto cleanup_errno;
-		}
-	}
-#endif
+	fcntl (fileno (db->fp), F_SETFD, FD_CLOEXEC);
 
 	buflen = BUFLEN;
 	buf = (char *) malloc (buflen);
@@ -543,10 +655,10 @@ int commonio_open (struct commonio_db *db, int mode)
 		goto cleanup_ENOMEM;
 	}
 
-	while (db->ops->fgets (buf, buflen, db->fp) == buf) {
+	while (db->ops->fgets (buf, (int) buflen, db->fp) == buf) {
 		while (   ((cp = strrchr (buf, '\n')) == NULL)
 		       && (feof (db->fp) == 0)) {
-			int len;
+			size_t len;
 
 			buflen += BUFLEN;
 			cp = (char *) realloc (buf, buflen);
@@ -555,7 +667,9 @@ int commonio_open (struct commonio_db *db, int mode)
 			}
 			buf = cp;
 			len = strlen (buf);
-			if (db->ops->fgets (buf + len, buflen - len, db->fp) == NULL) {
+			if (db->ops->fgets (buf + len,
+			                    (int) (buflen - len),
+			                    db->fp) == NULL) {
 				goto cleanup_buf;
 			}
 		}
@@ -619,12 +733,6 @@ int commonio_open (struct commonio_db *db, int mode)
       cleanup_errno:
 	saved_errno = errno;
 	free_linked_list (db);
-#ifdef WITH_SELINUX
-	if (db->scontext != NULL) {
-		freecon (db->scontext);
-		db->scontext = NULL;
-	}
-#endif
 	fclose (db->fp);
 	db->fp = NULL;
 	errno = saved_errno;
@@ -638,11 +746,27 @@ int
 commonio_sort (struct commonio_db *db, int (*cmp) (const void *, const void *))
 {
 	struct commonio_entry **entries, *ptr;
-	int n = 0, i;
+	size_t n = 0, i;
+#if KEEP_NIS_AT_END
+	struct commonio_entry *nis = NULL;
+#endif
 
-	for (ptr = db->head; NULL != ptr; ptr = ptr->next) {
+	for (ptr = db->head;
+	        (NULL != ptr)
+#if KEEP_NIS_AT_END
+	     && (NULL != ptr->line)
+	     && (   ('+' != ptr->line[0])
+	         && ('-' != ptr->line[0]))
+#endif
+	     ;
+	     ptr = ptr->next) {
 		n++;
 	}
+#if KEEP_NIS_AT_END
+	if ((NULL != ptr) && (NULL != ptr->line)) {
+		nis = ptr;
+	}
+#endif
 
 	if (n <= 1) {
 		return 0;
@@ -654,18 +778,40 @@ commonio_sort (struct commonio_db *db, int (*cmp) (const void *, const void *))
 	}
 
 	n = 0;
-	for (ptr = db->head; NULL != ptr; ptr = ptr->next) {
-		entries[n++] = ptr;
+	for (ptr = db->head;
+#if KEEP_NIS_AT_END
+	     nis != ptr;
+#else
+	     NULL != ptr;
+#endif
+/*@ -nullderef @*/
+	     ptr = ptr->next
+/*@ +nullderef @*/
+	    ) {
+		entries[n] = ptr;
+		n++;
 	}
 	qsort (entries, n, sizeof (struct commonio_entry *), cmp);
 
+	/* Take care of the head and tail separately */
 	db->head = entries[0];
-	db->tail = entries[--n];
+	n--;
+#if KEEP_NIS_AT_END
+	if (NULL == nis)
+#endif
+	{
+		db->tail = entries[n];
+	}
 	db->head->prev = NULL;
 	db->head->next = entries[1];
-	db->tail->prev = entries[n - 1];
-	db->tail->next = NULL;
+	entries[n]->prev = entries[n - 1];
+#if KEEP_NIS_AT_END
+	entries[n]->next = nis;
+#else
+	entries[n]->next = NULL;
+#endif
 
+	/* Now other elements have prev and next entries */
 	for (i = 1; i < n; i++) {
 		entries[i]->prev = entries[i - 1];
 		entries[i]->next = entries[i + 1];
@@ -680,7 +826,8 @@ commonio_sort (struct commonio_db *db, int (*cmp) (const void *, const void *))
 /*
  * Sort entries in db according to order in another.
  */
-int commonio_sort_wrt (struct commonio_db *shadow, struct commonio_db *passwd)
+int commonio_sort_wrt (struct commonio_db *shadow,
+                       const struct commonio_db *passwd)
 {
 	struct commonio_entry *head = NULL, *pw_ptr, *spw_ptr;
 	const char *name;
@@ -697,6 +844,9 @@ int commonio_sort_wrt (struct commonio_db *shadow, struct commonio_db *passwd)
 		for (spw_ptr = shadow->head;
 		     NULL != spw_ptr;
 		     spw_ptr = spw_ptr->next) {
+			if (NULL == spw_ptr->eptr) {
+				continue;
+			}
 			if (strcmp (name, shadow->ops->getname (spw_ptr->eptr))
 			    == 0) {
 				break;
@@ -732,6 +882,7 @@ int commonio_sort_wrt (struct commonio_db *shadow, struct commonio_db *passwd)
  * It returns 0 if all the entries could be written correctly.
  */
 static int write_all (const struct commonio_db *db)
+	/*@requires notnull db->fp@*/
 {
 	const struct commonio_entry *p;
 	void *eptr;
@@ -739,6 +890,7 @@ static int write_all (const struct commonio_db *db)
 	for (p = db->head; NULL != p; p = p->next) {
 		if (p->changed) {
 			eptr = p->eptr;
+			assert (NULL != eptr);
 			if (db->ops->put (eptr, db->fp) != 0) {
 				return -1;
 			}
@@ -756,6 +908,7 @@ static int write_all (const struct commonio_db *db)
 
 
 int commonio_close (struct commonio_db *db)
+	/*@requires notnull db->fp@*/
 {
 	char buf[1024];
 	int errors = 0;
@@ -768,7 +921,7 @@ int commonio_close (struct commonio_db *db)
 	db->isopen = false;
 
 	if (!db->changed || db->readonly) {
-		fclose (db->fp);
+		(void) fclose (db->fp);
 		db->fp = NULL;
 		goto success;
 	}
@@ -780,27 +933,21 @@ int commonio_close (struct commonio_db *db)
 	memzero (&sb, sizeof sb);
 	if (NULL != db->fp) {
 		if (fstat (fileno (db->fp), &sb) != 0) {
-			fclose (db->fp);
+			(void) fclose (db->fp);
 			db->fp = NULL;
 			goto fail;
 		}
-#ifdef WITH_SELINUX
-		if (db->scontext != NULL) {
-			if (getfscreatecon (&old_context) < 0) {
-				errors++;
-				goto fail;
-			}
-			if (setfscreatecon (db->scontext) < 0) {
-				errors++;
-				goto fail;
-			}
-		}
-#endif
+
 		/*
 		 * Create backup file.
 		 */
 		snprintf (buf, sizeof buf, "%s-", db->filename);
 
+#ifdef WITH_SELINUX
+		if (set_selinux_file_context (buf) != 0) {
+			errors++;
+		}
+#endif
 		if (create_backup (buf, db->fp) != 0) {
 			errors++;
 		}
@@ -809,6 +956,11 @@ int commonio_close (struct commonio_db *db)
 			errors++;
 		}
 
+#ifdef WITH_SELINUX
+		if (reset_selinux_file_context () != 0) {
+			errors++;
+		}
+#endif
 		if (errors != 0) {
 			db->fp = NULL;
 			goto fail;
@@ -841,9 +993,9 @@ int commonio_close (struct commonio_db *db)
 	if (fsync (fileno (db->fp)) != 0) {
 		errors++;
 	}
-#else
+#else				/* !HAVE_FSYNC */
 	sync ();
-#endif
+#endif				/* !HAVE_FSYNC */
 	if (fclose (db->fp) != 0) {
 		errors++;
 	}
@@ -865,26 +1017,14 @@ int commonio_close (struct commonio_db *db)
 	errors++;
       success:
 
-#ifdef WITH_SELINUX
-	if (db->scontext != NULL) {
-		if (setfscreatecon (old_context) < 0) {
-			errors++;
-		}
-		if (NULL != old_context) {
-			freecon (old_context);
-			old_context = NULL;
-		}
-		freecon (db->scontext);
-		db->scontext = NULL;
-	}
-#endif
 	free_linked_list (db);
 	return errors == 0;
 }
 
-static struct commonio_entry *next_entry_by_name (struct commonio_db *db,
-                                                  struct commonio_entry *pos,
-                                                  const char *name)
+static /*@dependent@*/ /*@null@*/struct commonio_entry *next_entry_by_name (
+	struct commonio_db *db,
+	/*@null@*/struct commonio_entry *pos,
+	const char *name)
 {
 	struct commonio_entry *p;
 	void *ep;
@@ -903,10 +1043,11 @@ static struct commonio_entry *next_entry_by_name (struct commonio_db *db,
 	return p;
 }
 
-static struct commonio_entry *find_entry_by_name (struct commonio_db *db,
-						  const char *name)
+static /*@dependent@*/ /*@null@*/struct commonio_entry *find_entry_by_name (
+	struct commonio_db *db,
+	const char *name)
 {
-	return next_entry_by_name(db, db->head, name);
+	return next_entry_by_name (db, db->head, name);
 }
 
 
@@ -952,9 +1093,9 @@ int commonio_update (struct commonio_db *db, const void *eptr)
 
 #if KEEP_NIS_AT_END
 	add_one_entry_nis (db, p);
-#else
+#else				/* !KEEP_NIS_AT_END */
 	add_one_entry (db, p);
-#endif
+#endif				/* !KEEP_NIS_AT_END */
 
 	db->changed = true;
 	return 1;
@@ -1025,7 +1166,7 @@ int commonio_remove (struct commonio_db *db, const char *name)
  *
  *	Otherwise, it returns NULL.
  */
-const void *commonio_locate (struct commonio_db *db, const char *name)
+/*@observer@*/ /*@null@*/const void *commonio_locate (struct commonio_db *db, const char *name)
 {
 	struct commonio_entry *p;
 
@@ -1062,7 +1203,7 @@ int commonio_rewind (struct commonio_db *db)
  *
  * It returns the next entry, or NULL if no other entries could be found.
  */
-const void *commonio_next (struct commonio_db *db)
+/*@observer@*/ /*@null@*/const void *commonio_next (struct commonio_db *db)
 {
 	void *eptr;
 
