@@ -2,7 +2,7 @@
  * Copyright (c) 1989 - 1994, Julianne Frances Haugh
  * Copyright (c) 1996 - 2001, Marek Michałkiewicz
  * Copyright (c) 2001 - 2006, Tomasz Kłoczko
- * Copyright (c) 2007 - 2009, Nicolas François
+ * Copyright (c) 2007 - 2011, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -32,7 +32,7 @@
 
 #include <config.h>
 
-#ident "$Id: login.c 2866 2009-05-09 13:14:50Z nekral-guest $"
+#ident "$Id: login.c 3549 2011-11-06 18:38:51Z nekral-guest $"
 
 #include <errno.h>
 #include <grp.h>
@@ -82,7 +82,7 @@ static pam_handle_t *pamh = NULL;
 /*
  * Global variables
  */
-char *Prog;
+const char *Prog;
 
 static const char *hostname = "";
 static /*@null@*/ /*@only@*/char *username = NULL;
@@ -126,10 +126,12 @@ static void usage (void);
 static void setup_tty (void);
 static void process_flags (int argc, char *const *argv);
 static /*@observer@*/const char *get_failent_user (/*@returned@*/const char *user);
+#ifndef USE_PAM
 static void update_utmp (const char *user,
                          const char *tty,
                          const char *host,
                          /*@null@*/const struct utmp *utent);
+#endif				/* ! USE_PAM */
 
 #ifndef USE_PAM
 static struct faillog faillog;
@@ -167,10 +169,10 @@ static void usage (void)
 static void setup_tty (void)
 {
 	TERMIO termio;
-	int erasechar;
-	int killchar;
 
 	if (GTTY (0, &termio) == 0) {	/* get terminal characteristics */
+		int erasechar;
+		int killchar;
 
 		/*
 		 * Add your favorite terminal modes here ...
@@ -242,7 +244,6 @@ static void check_nologin (bool login_to_root)
 	fname = getdef_str ("NOLOGINS_FILE");
 	if ((NULL != fname) && (access (fname, F_OK) == 0)) {
 		FILE *nlfp;
-		int c;
 
 		/*
 		 * Cat the file if it can be opened, otherwise just
@@ -250,6 +251,7 @@ static void check_nologin (bool login_to_root)
 		 */
 		nlfp = fopen (fname, "r");
 		if (NULL != nlfp) {
+			int c;
 			while ((c = getc (nlfp)) != EOF) {
 				if (c == '\n') {
 					(void) putchar ('\r');
@@ -471,6 +473,7 @@ static /*@observer@*/const char *get_failent_user (/*@returned@*/const char *use
 	return failent_user;
 }
 
+#ifndef USE_PAM
 /*
  * update_utmp - Update or create an utmp entry in utmp, wtmp, utmpw, and
  *               wtmpx
@@ -496,6 +499,7 @@ static void update_utmp (const char *user,
 	free (utx);
 #endif				/* USE_UTMPX */
 }
+#endif				/* ! USE_PAM */
 
 /*
  * login - create a new login session for a user
@@ -527,14 +531,13 @@ int main (int argc, char **argv)
 #endif
 	unsigned int delay;
 	unsigned int retries;
-	bool failed;
 	bool subroot = false;
 #ifndef USE_PAM
 	bool is_console;
 #endif
 	int err;
 	const char *cp;
-	char *tmp;
+	const char *tmp;
 	char fromhost[512];
 	struct passwd *pwd = NULL;
 	char **envp = environ;
@@ -773,7 +776,7 @@ int main (int argc, char **argv)
 		 */
 		failcount = 0;
 		while (true) {
-			failed = false;
+			bool failed = false;
 
 			failcount++;
 #ifdef HAS_PAM_FAIL_DELAY
@@ -792,9 +795,9 @@ int main (int argc, char **argv)
 				SYSLOG ((LOG_NOTICE,
 				         "TOO MANY LOGIN TRIES (%u)%s FOR '%s'",
 				         failcount, fromhost, failent_user));
-				fprintf(stderr,
-				        _("Maximum number of tries exceeded (%u)\n"),
-				        failcount);
+				fprintf (stderr,
+				         _("Maximum number of tries exceeded (%u)\n"),
+				         failcount);
 				PAM_END;
 				exit(0);
 			} else if (retcode == PAM_ABORT) {
@@ -836,9 +839,9 @@ int main (int argc, char **argv)
 				SYSLOG ((LOG_NOTICE,
 				         "TOO MANY LOGIN TRIES (%u)%s FOR '%s'",
 				         failcount, fromhost, failent_user));
-				fprintf(stderr,
-				        _("Maximum number of tries exceeded (%u)\n"),
-				        failcount);
+				fprintf (stderr,
+				         _("Maximum number of tries exceeded (%u)\n"),
+				         failcount);
 				PAM_END;
 				exit(0);
 			}
@@ -878,12 +881,15 @@ int main (int argc, char **argv)
 	if (NULL != username) {
 		free (username);
 	}
-	username = pam_user;
+	username = xstrdup (pam_user);
 	failent_user = get_failent_user (username);
 
 	pwd = xgetpwnam (username);
 	if (NULL == pwd) {
 		SYSLOG ((LOG_ERR, "cannot find user %s", failent_user));
+		fprintf (stderr,
+		         _("Cannot find user (%s)\n"),
+		         username);
 		exit (1);
 	}
 
@@ -903,6 +909,7 @@ int main (int argc, char **argv)
 
 #else				/* ! USE_PAM */
 	while (true) {	/* repeatedly get login/password pairs */
+		bool failed;
 		/* user_passwd is always a pointer to this constant string
 		 * or a passwd or shadow password that will be memzero by
 		 * pw_free / spw_free.
@@ -913,6 +920,7 @@ int main (int argc, char **argv)
 		 * anymore. */
 		if (NULL != pwd) {
 			pw_free (pwd);
+			pwd = NULL;
 		}
 		if (NULL != spwd) {
 			spw_free (spwd);
@@ -1207,11 +1215,13 @@ int main (int argc, char **argv)
 		}
 	}
 
+#ifndef USE_PAM
 	/*
 	 * The utmp entry needs to be updated to indicate the new status
 	 * of the session, the new PID and SID.
 	 */
 	update_utmp (username, tty, hostname, utent);
+#endif				/* ! USE_PAM */
 
 	/* The pwd and spwd entries for the user have been copied.
 	 *
