@@ -2,7 +2,7 @@
  * Copyright (c) 1991 - 1994, Julianne Frances Haugh
  * Copyright (c) 1996 - 2000, Marek Michałkiewicz
  * Copyright (c) 2000 - 2006, Tomasz Kłoczko
- * Copyright (c) 2007 - 2008, Nicolas François
+ * Copyright (c) 2007 - 2011, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -32,7 +32,7 @@
 
 #include <config.h>
 
-#ident "$Id: userdel.c 2874 2009-05-09 13:15:44Z nekral-guest $"
+#ident "$Id: userdel.c 3640 2011-11-19 21:51:52Z nekral-guest $"
 
 #include <errno.h>
 #include <fcntl.h>
@@ -58,7 +58,11 @@
 #include "shadowio.h"
 #ifdef	SHADOWGRP
 #include "sgroupio.h"
-#endif
+#endif				/* SHADOWGRP */
+#ifdef WITH_TCB
+#include <tcb.h>
+#include "tcbfuncs.h"
+#endif				/* WITH_TCB */
 /*@-exitarg@*/
 #include "exitcodes.h"
 
@@ -70,59 +74,73 @@
 #define E_USER_BUSY	8	/* user currently logged in */
 #define E_GRP_UPDATE	10	/* can't update group file */
 #define E_HOMEDIR	12	/* can't remove home directory */
+#define E_SE_UPDATE	14	/* can't update SELinux user mapping */
 
 /*
  * Global variables
  */
-char *Prog;
+const char *Prog;
 
 static char *user_name;
 static uid_t user_id;
+static gid_t user_gid;
 static char *user_home;
 
 static bool fflg = false;
 static bool rflg = false;
+static bool Zflg = false;
 
 static bool is_shadow_pwd;
 
 #ifdef SHADOWGRP
 static bool is_shadow_grp;
 static bool sgr_locked = false;
-#endif
+#endif				/* SHADOWGRP */
 static bool pw_locked  = false;
 static bool gr_locked   = false;
 static bool spw_locked  = false;
 
 /* local function prototypes */
-static void usage (void);
+static void usage (int status);
 static void update_groups (void);
+static void remove_usergroup (void);
 static void close_files (void);
 static void fail_exit (int);
 static void open_files (void);
 static void update_user (void);
-static void user_busy (const char *, uid_t);
 static void user_cancel (const char *);
 
 #ifdef EXTRA_CHECK_HOME_DIR
 static bool path_prefix (const char *, const char *);
-#endif
+#endif				/* EXTRA_CHECK_HOME_DIR */
 static int is_owner (uid_t, const char *);
-static void remove_mailbox (void);
+static int remove_mailbox (void);
+#ifdef WITH_TCB
+static int remove_tcbdir (const char *user_name, uid_t user_id);
+#endif				/* WITH_TCB */
 
 /*
  * usage - display usage message and exit
  */
-static void usage (void)
+static void usage (int status)
 {
-	fputs (_("Usage: userdel [options] LOGIN\n"
-	         "\n"
-	         "Options:\n"
-	         "  -f, --force                   force removal of files,\n"
-	         "                                even if not owned by user\n"
-	         "  -h, --help                    display this help message and exit\n"
-	         "  -r, --remove                  remove home directory and mail spool\n"
-	         "\n"), stderr);
-	exit (E_USAGE);
+	FILE *usageout = (E_SUCCESS != status) ? stderr : stdout;
+	(void) fprintf (usageout,
+	                _("Usage: %s [options] LOGIN\n"
+	                  "\n"
+	                  "Options:\n"),
+	                Prog);
+	(void) fputs (_("  -f, --force                   force removal of files,\n"
+	                "                                even if not owned by user\n"),
+	              usageout);
+	(void) fputs (_("  -h, --help                    display this help message and exit\n"), usageout);
+	(void) fputs (_("  -r, --remove                  remove home directory and mail spool\n"), usageout);
+	(void) fputs (_("  -R, --root CHROOT_DIR         directory to chroot into\n"), usageout);
+#ifdef WITH_SELINUX
+	(void) fputs (_("  -Z, --selinux-user            remove any SELinux user mapping for the user\n"), usageout);
+#endif				/* WITH_SELINUX */
+	(void) fputs ("\n", usageout);
+	exit (status);
 }
 
 /*
@@ -139,10 +157,8 @@ static void update_groups (void)
 {
 	const struct group *grp;
 	struct group *ngrp;
-	struct passwd *pwd;
 
 #ifdef	SHADOWGRP
-	bool deleted_user_group = false;
 	const struct sgrp *sgrp;
 	struct sgrp *nsgrp;
 #endif				/* SHADOWGRP */
@@ -161,7 +177,7 @@ static void update_groups (void)
 			continue;
 		}
 
-		/* 
+		/*
 		 * Delete the username from the list of group members and
 		 * update the group entry to reflect the change.
 		 */
@@ -188,69 +204,15 @@ static void update_groups (void)
 		              "deleting user from group",
 		              user_name, (unsigned int) user_id,
 		              SHADOW_AUDIT_SUCCESS);
-#endif
+#endif				/* WITH_AUDIT */
 		SYSLOG ((LOG_INFO, "delete '%s' from group '%s'\n",
 			 user_name, ngrp->gr_name));
 	}
 
-	/*
-	 * we've removed their name from all the groups above, so
-	 * now if they have a group with the same name as their
-	 * user name, with no members, we delete it.
-	 */
-	grp = xgetgrnam (user_name);
-	if (   (NULL != grp)
-	    && getdef_bool ("USERGROUPS_ENAB")
-	    && (NULL == grp->gr_mem[0])) {
-
-		pwd = NULL;
-		if (!fflg) {
-			/*
-			 * Scan the passwd file to check if this group is still
-			 * used as a primary group.
-			 */
-			setpwent ();
-			while ((pwd = getpwent ()) != NULL) {
-				if (strcmp (pwd->pw_name, user_name) == 0) {
-					continue;
-				}
-				if (pwd->pw_gid == grp->gr_gid) {
-					fprintf (stderr,
-					         _("%s: group %s is the primary group of another user and is not removed.\n"),
-					         Prog, grp->gr_name);
-					break;
-				}
-			}
-			endpwent ();
-		}
-
-		if (NULL == pwd) {
-			/*
-			 * We can remove this group, it is not the primary
-			 * group of any remaining user.
-			 */
-			if (gr_remove (grp->gr_name) == 0) {
-				fprintf (stderr,
-				         _("%s: cannot remove entry '%s' from %s\n"),
-				         Prog, grp->gr_name, gr_dbname ());
-				fail_exit (E_GRP_UPDATE);
-			}
-
-#ifdef SHADOWGRP
-			deleted_user_group = true;
-#endif
-
-#ifdef WITH_AUDIT
-			audit_logger (AUDIT_DEL_GROUP, Prog,
-			              "deleting group",
-			              grp->gr_name, AUDIT_NO_ID,
-			              SHADOW_AUDIT_SUCCESS);
-#endif
-			SYSLOG ((LOG_INFO,
-				 "removed group '%s' owned by '%s'\n",
-				 grp->gr_name, user_name));
-		}
+	if (getdef_bool ("USERGROUPS_ENAB")) {
+		remove_usergroup ();
 	}
+
 #ifdef	SHADOWGRP
 	if (!is_shadow_grp) {
 		return;
@@ -304,23 +266,113 @@ static void update_groups (void)
 		              "deleting user from shadow group",
 		              user_name, (unsigned int) user_id,
 		              SHADOW_AUDIT_SUCCESS);
-#endif
+#endif				/* WITH_AUDIT */
 		SYSLOG ((LOG_INFO, "delete '%s' from shadow group '%s'\n",
-			 user_name, nsgrp->sg_name));
-	}
-
-	if (   deleted_user_group
-	    && (sgr_locate (user_name) != NULL)) {
-		if (sgr_remove (user_name) == 0) {
-			fprintf (stderr,
-			         _("%s: cannot remove entry '%s' from %s\n"),
-			         Prog, user_name, sgr_dbname ());
-			fail_exit (E_GRP_UPDATE);
-		}
+		         user_name, nsgrp->sg_name));
 	}
 #endif				/* SHADOWGRP */
 }
 
+/*
+ * remove_usergroup - delete the user's group if it is a usergroup
+ *
+ *	An usergroup is removed if
+ *	  + it has the same name as the user
+ *	  + it is the primary group of the user
+ *	  + it has no other members
+ *	  + it is not the primary group of any other user
+ */
+static void remove_usergroup (void)
+{
+	const struct group *grp;
+	const struct passwd *pwd = NULL;
+
+	grp = gr_locate (user_name);
+	if (NULL == grp) {
+		/* This user has no usergroup. */
+		return;
+	}
+
+	if (grp->gr_gid != user_gid) {
+		fprintf (stderr,
+		         _("%s: group %s not removed because it is not the primary group of user %s.\n"),
+		         Prog, grp->gr_name, user_name);
+		return;
+	}
+
+	if (NULL != grp->gr_mem[0]) {
+		/* The usergroup has other members. */
+		fprintf (stderr,
+		         _("%s: group %s not removed because it has other members.\n"),
+		         Prog, grp->gr_name);
+		return;
+	}
+
+	if (!fflg) {
+		/*
+		 * Scan the passwd file to check if this group is still
+		 * used as a primary group.
+		 */
+		setpwent ();
+		while ((pwd = getpwent ()) != NULL) {
+			if (strcmp (pwd->pw_name, user_name) == 0) {
+				continue;
+			}
+			if (pwd->pw_gid == grp->gr_gid) {
+				fprintf (stderr,
+				         _("%s: group %s is the primary group of another user and is not removed.\n"),
+				         Prog, grp->gr_name);
+				break;
+			}
+		}
+		endpwent ();
+	}
+
+	if (NULL == pwd) {
+		/*
+		 * We can remove this group, it is not the primary
+		 * group of any remaining user.
+		 */
+		if (gr_remove (grp->gr_name) == 0) {
+			fprintf (stderr,
+			         _("%s: cannot remove entry '%s' from %s\n"),
+			         Prog, grp->gr_name, gr_dbname ());
+			fail_exit (E_GRP_UPDATE);
+		}
+
+#ifdef WITH_AUDIT
+		audit_logger (AUDIT_DEL_GROUP, Prog,
+		              "deleting group",
+		              grp->gr_name, AUDIT_NO_ID,
+		              SHADOW_AUDIT_SUCCESS);
+#endif				/* WITH_AUDIT */
+		SYSLOG ((LOG_INFO,
+		         "removed group '%s' owned by '%s'\n",
+		         grp->gr_name, user_name));
+
+#ifdef	SHADOWGRP
+		if (sgr_locate (user_name) != NULL) {
+			if (sgr_remove (user_name) == 0) {
+				fprintf (stderr,
+				         _("%s: cannot remove entry '%s' from %s\n"),
+				         Prog, user_name, sgr_dbname ());
+				fail_exit (E_GRP_UPDATE);
+			}
+#ifdef WITH_AUDIT
+			audit_logger (AUDIT_DEL_GROUP, Prog,
+			              "deleting shadow group",
+			              grp->gr_name, AUDIT_NO_ID,
+			              SHADOW_AUDIT_SUCCESS);
+#endif				/* WITH_AUDIT */
+			SYSLOG ((LOG_INFO,
+			         "removed shadow group '%s' owned by '%s'\n",
+			         grp->gr_name, user_name));
+
+		}
+#endif				/* SHADOWGRP */
+	}
+}
+
 /*
  * close_files - close all of the files that were opened
  *
@@ -384,7 +436,7 @@ static void close_files (void)
 		}
 		sgr_locked = false;
 	}
-#endif
+#endif				/* SHADOWGRP */
 }
 
 /*
@@ -421,14 +473,14 @@ static void fail_exit (int code)
 			/* continue */
 		}
 	}
-#endif
+#endif				/* SHADOWGRP */
 
 #ifdef WITH_AUDIT
 	audit_logger (AUDIT_DEL_USER, Prog,
 	              "deleting user",
 	              user_name, (unsigned int) user_id,
 	              SHADOW_AUDIT_FAILURE);
-#endif
+#endif				/* WITH_AUDIT */
 
 	exit (code);
 }
@@ -450,7 +502,7 @@ static void open_files (void)
 		              "locking password file",
 		              user_name, (unsigned int) user_id,
 		              SHADOW_AUDIT_FAILURE);
-#endif
+#endif				/* WITH_AUDIT */
 		fail_exit (E_PW_UPDATE);
 	}
 	pw_locked = true;
@@ -462,7 +514,7 @@ static void open_files (void)
 		              "opening password file",
 		              user_name, (unsigned int) user_id,
 		              SHADOW_AUDIT_FAILURE);
-#endif
+#endif				/* WITH_AUDIT */
 		fail_exit (E_PW_UPDATE);
 	}
 	if (is_shadow_pwd) {
@@ -475,7 +527,7 @@ static void open_files (void)
 			              "locking shadow password file",
 			              user_name, (unsigned int) user_id,
 			              SHADOW_AUDIT_FAILURE);
-#endif
+#endif				/* WITH_AUDIT */
 			fail_exit (E_PW_UPDATE);
 		}
 		spw_locked = true;
@@ -488,7 +540,7 @@ static void open_files (void)
 			              "opening shadow password file",
 			              user_name, (unsigned int) user_id,
 			              SHADOW_AUDIT_FAILURE);
-#endif
+#endif				/* WITH_AUDIT */
 			fail_exit (E_PW_UPDATE);
 		}
 	}
@@ -501,7 +553,7 @@ static void open_files (void)
 		              "locking group file",
 		              user_name, (unsigned int) user_id,
 		              SHADOW_AUDIT_FAILURE);
-#endif
+#endif				/* WITH_AUDIT */
 		fail_exit (E_GRP_UPDATE);
 	}
 	gr_locked = true;
@@ -512,7 +564,7 @@ static void open_files (void)
 		              "opening group file",
 		              user_name, (unsigned int) user_id,
 		              SHADOW_AUDIT_FAILURE);
-#endif
+#endif				/* WITH_AUDIT */
 		fail_exit (E_GRP_UPDATE);
 	}
 #ifdef	SHADOWGRP
@@ -526,7 +578,7 @@ static void open_files (void)
 			              "locking shadow group file",
 			              user_name, (unsigned int) user_id,
 			              SHADOW_AUDIT_FAILURE);
-#endif
+#endif				/* WITH_AUDIT */
 			fail_exit (E_GRP_UPDATE);
 		}
 		sgr_locked= true;
@@ -538,11 +590,11 @@ static void open_files (void)
 			              "opening shadow group file",
 			              user_name, (unsigned int) user_id,
 			              SHADOW_AUDIT_FAILURE);
-#endif
+#endif				/* WITH_AUDIT */
 			fail_exit (E_GRP_UPDATE);
 		}
 	}
-#endif
+#endif				/* SHADOWGRP */
 }
 
 /*
@@ -572,115 +624,31 @@ static void update_user (void)
 	              "deleting user entries",
 	              user_name, (unsigned int) user_id,
 	              SHADOW_AUDIT_SUCCESS);
-#endif
+#endif				/* WITH_AUDIT */
 	SYSLOG ((LOG_INFO, "delete user '%s'\n", user_name));
 }
 
 /*
- * user_busy - see if user is logged in.
- *
- * XXX - should probably check if there are any processes owned
- * by this user. Also, I think this check should be in usermod
- * as well (at least when changing username or UID).  --marekm
- */
-static void user_busy (const char *name, uid_t uid)
-{
-
-/*
- * We see if the user is logged in by looking for the user name
- * in the utmp file.
- */
-#ifdef USE_UTMPX
-	struct utmpx *utent;
-
-	setutxent ();
-	while ((utent = getutxent ()) != NULL)
-#else				/* !USE_UTMPX */
-	struct utmp *utent;
-
-	setutent ();
-	while ((utent = getutent ()) != NULL)
-#endif				/* !USE_UTMPX */
-	{
-		if (utent->ut_type != USER_PROCESS)
-			continue;
-
-		if (strncmp (utent->ut_user, name, sizeof utent->ut_user) != 0) {
-			continue;
-		}
-		fprintf (stderr,
-			 _("%s: user %s is currently logged in\n"), Prog, name);
-		if (!fflg) {
-#ifdef WITH_AUDIT
-			audit_logger (AUDIT_DEL_USER, Prog,
-			              "deleting user logged in",
-			              name, AUDIT_NO_ID,
-			              SHADOW_AUDIT_FAILURE);
-#endif
-			exit (E_USER_BUSY);
-		}
-	}
-}
-
-/* 
  * user_cancel - cancel cron and at jobs
  *
- *	user_cancel removes the crontab and any at jobs for a user
+ *	user_cancel calls a script for additional cleanups like removal of
+ *	cron, at, or print jobs.
  */
 
-/*
- * We used to have all this stuff hardcoded here, but now
- * we just run an external script - it may need to do other
- * things as well (like removing print jobs) and we may not
- * want to recompile userdel too often. Below is a sample
- * script (should work at least on Debian 1.1).  --marekm
-==========
-#! /bin/sh
-
-# Check for the required argument.
-if [ $# != 1 ]; then
-	echo Usage: $0 username
-	exit 1
-fi
-
-# Remove cron jobs.
-crontab -r -u $1
-
-# Remove at jobs. XXX - will remove any jobs owned by the same UID, even if
-# it was shared by a different username. at really should store the username
-# somewhere, and atrm should support an option to remove all jobs owned by
-# the specified user - for now we have to do this ugly hack...
-find /var/spool/cron/atjobs -name "[^.]*" -type f -user $1 -exec rm {} \;
-
-# Remove print jobs.
-lprm $1
-
-# All done.
-exit 0
-==========
- */
 static void user_cancel (const char *user)
 {
-	char *cmd;
-	pid_t pid, wpid;
+	const char *cmd;
+	const char *argv[3];
 	int status;
 
 	cmd = getdef_str ("USERDEL_CMD");
 	if (NULL == cmd) {
 		return;
 	}
-	pid = fork ();
-	if (pid == 0) {
-		execl (cmd, cmd, user, (char *) 0);
-		perror (cmd);
-		_exit (errno == ENOENT ? E_CMD_NOTFOUND : E_CMD_NOEXEC);
-	} else if ((pid_t)-1 == pid) {
-		perror ("fork");
-		return;
-	}
-	do {
-		wpid = wait (&status);
-	} while ((wpid != pid) && ((pid_t)-1 != wpid));
+	argv[0] = cmd;
+	argv[1] = user;
+	argv[2] = (char *)0;
+	(void) run_command (cmd, argv, NULL, &status);
 }
 
 #ifdef EXTRA_CHECK_HOME_DIR
@@ -690,44 +658,84 @@ static bool path_prefix (const char *s1, const char *s2)
 	        && (   ('\0' == s2[strlen (s1)])
 	            || ('/'  == s2[strlen (s1)])));
 }
-#endif
+#endif				/* EXTRA_CHECK_HOME_DIR */
 
+/*
+ * is_owner - Check if path is owned by uid
+ *
+ * Return
+ *  1: path exists and is owned by uid
+ *  0: path is not owned by uid, or a failure occurred
+ * -1: path does not exist
+ */
 static int is_owner (uid_t uid, const char *path)
 {
 	struct stat st;
 
+	errno = 0;
 	if (stat (path, &st) != 0) {
-		return -1;
+		if ((ENOENT == errno) || (ENOTDIR == errno)) {
+			/* The file or directory does not exist */
+			return -1;
+		} else {
+			return 0;
+		}
 	}
-	return (st.st_uid == uid);
+	return (st.st_uid == uid) ? 1 : 0;
 }
 
-static void remove_mailbox (void)
+static int remove_mailbox (void)
 {
 	const char *maildir;
 	char mailfile[1024];
 	int i;
+	int errors = 0;
 
 	maildir = getdef_str ("MAIL_DIR");
 #ifdef MAIL_SPOOL_DIR
 	if ((NULL == maildir) && (getdef_str ("MAIL_FILE") == NULL)) {
 		maildir = MAIL_SPOOL_DIR;
 	}
-#endif
+#endif				/* MAIL_SPOOL_DIR */
 	if (NULL == maildir) {
-		return;
+		return 0;
 	}
 	snprintf (mailfile, sizeof mailfile, "%s/%s", maildir, user_name);
-	if (fflg) {
-		if (unlink (mailfile) != 0) {
-			fprintf (stderr, _("%s: warning: can't remove %s: %s"), Prog, mailfile, strerror (errno));
+
+	if (access (mailfile, F_OK) != 0) {
+		if (ENOENT == errno) {
+			fprintf (stderr,
+			         _("%s: %s mail spool (%s) not found\n"),
+			         Prog, user_name, mailfile);
+			return 0;
+		} else {
+			fprintf (stderr,
+			         _("%s: warning: can't remove %s: %s\n"),
+			         Prog, mailfile, strerror (errno));
 			SYSLOG ((LOG_ERR, "Cannot remove %s: %s", mailfile, strerror (errno)));
 #ifdef WITH_AUDIT
 			audit_logger (AUDIT_DEL_USER, Prog,
 			              "deleting mail file",
 			              user_name, (unsigned int) user_id,
 			              SHADOW_AUDIT_FAILURE);
-#endif
+#endif				/* WITH_AUDIT */
+			return -1;
+		}
+	}
+
+	if (fflg) {
+		if (unlink (mailfile) != 0) {
+			fprintf (stderr,
+			         _("%s: warning: can't remove %s: %s\n"),
+			         Prog, mailfile, strerror (errno));
+			SYSLOG ((LOG_ERR, "Cannot remove %s: %s", mailfile, strerror (errno)));
+#ifdef WITH_AUDIT
+			audit_logger (AUDIT_DEL_USER, Prog,
+			              "deleting mail file",
+			              user_name, (unsigned int) user_id,
+			              SHADOW_AUDIT_FAILURE);
+#endif				/* WITH_AUDIT */
+			errors = 1;
 			/* continue */
 		}
 #ifdef WITH_AUDIT
@@ -738,34 +746,39 @@ static void remove_mailbox (void)
 			              user_name, (unsigned int) user_id,
 			              SHADOW_AUDIT_SUCCESS);
 		}
-#endif
-		return;
+#endif				/* WITH_AUDIT */
+		return errors;
 	}
 	i = is_owner (user_id, mailfile);
 	if (i == 0) {
 		fprintf (stderr,
 		         _("%s: %s not owned by %s, not removing\n"),
 		         Prog, mailfile, user_name);
-		SYSLOG ((LOG_ERR, "%s not owned by %s, not removed", mailfile, strerror (errno)));
+		SYSLOG ((LOG_ERR,
+		         "%s not owned by %s, not removed",
+		         mailfile, strerror (errno)));
 #ifdef WITH_AUDIT
 		audit_logger (AUDIT_DEL_USER, Prog,
 		              "deleting mail file",
 		              user_name, (unsigned int) user_id,
 		              SHADOW_AUDIT_FAILURE);
-#endif
-		return;
+#endif				/* WITH_AUDIT */
+		return 1;
 	} else if (i == -1) {
-		return;		/* mailbox doesn't exist */
+		return 0;		/* mailbox doesn't exist */
 	}
 	if (unlink (mailfile) != 0) {
-		fprintf (stderr, _("%s: warning: can't remove %s: %s"), Prog, mailfile, strerror (errno));
+		fprintf (stderr,
+		         _("%s: warning: can't remove %s: %s\n"),
+		         Prog, mailfile, strerror (errno));
 		SYSLOG ((LOG_ERR, "Cannot remove %s: %s", mailfile, strerror (errno)));
 #ifdef WITH_AUDIT
 		audit_logger (AUDIT_DEL_USER, Prog,
 		              "deleting mail file",
 		              user_name, (unsigned int) user_id,
 		              SHADOW_AUDIT_FAILURE);
-#endif
+#endif				/* WITH_AUDIT */
+		errors = 1;
 		/* continue */
 	}
 #ifdef WITH_AUDIT
@@ -776,9 +789,57 @@ static void remove_mailbox (void)
 		              user_name, (unsigned int) user_id,
 		              SHADOW_AUDIT_SUCCESS);
 	}
-#endif
+#endif				/* WITH_AUDIT */
+	return errors;
 }
 
+#ifdef WITH_TCB
+static int remove_tcbdir (const char *user_name, uid_t user_id)
+{
+	char *buf;
+	int ret = 0;
+	size_t bufsize = (sizeof TCB_DIR) + strlen (user_name) + 2;
+
+	if (!getdef_bool ("USE_TCB")) {
+		return 0;
+	}
+
+	buf = malloc (buflen);
+	if (NULL == buf) {
+		fprintf (stderr, _("%s: Can't allocate memory, "
+		                   "tcb entry for %s not removed.\n"),
+		         Prog, user_name);
+		return 1;
+	}
+	snprintf (buf, buflen, TCB_DIR "/%s", user_name);
+	if (shadowtcb_drop_priv () == SHADOWTCB_FAILURE) {
+		fprintf (stderr, _("%s: Cannot drop privileges: %s\n"),
+		         Prog, strerror (errno));
+		shadowtcb_gain_priv ();
+		free (buf);
+		return 1;
+	}
+	/* Only remove directory contents with dropped privileges.
+	 * We will regain them and remove the user's tcb directory afterwards.
+	 */
+	if (remove_tree (buf, false) != 0) {
+		fprintf (stderr, _("%s: Cannot remove the content of %s: %s\n"),
+		         Prog, buf, strerror (errno));
+		shadowtcb_gain_priv ();
+		free (buf);
+		return 1;
+	}
+	shadowtcb_gain_priv ();
+	free (buf);
+	if (shadowtcb_remove (user_name) == SHADOWTCB_FAILURE) {
+		fprintf (stderr, _("%s: Cannot remove tcb files for %s: %s\n"),
+		         Prog, user_name, strerror (errno));
+		ret = 1;
+	}
+	return ret;
+}
+#endif				/* WITH_TCB */
+
 /*
  * main - userdel command
  */
@@ -793,10 +854,6 @@ int main (int argc, char **argv)
 #endif				/* USE_PAM */
 #endif				/* ACCT_TOOLS_SETUID */
 
-#ifdef WITH_AUDIT
-	audit_help_open ();
-#endif
-
 	/*
 	 * Get my name so that I can use it to report errors.
 	 */
@@ -805,38 +862,70 @@ int main (int argc, char **argv)
 	(void) bindtextdomain (PACKAGE, LOCALEDIR);
 	(void) textdomain (PACKAGE);
 
+	process_root_flag ("-R", argc, argv);
+
+	OPENLOG ("userdel");
+#ifdef WITH_AUDIT
+	audit_help_open ();
+#endif				/* WITH_AUDIT */
+
 	{
 		/*
 		 * Parse the command line options.
 		 */
 		int c;
 		static struct option long_options[] = {
-			{"force", no_argument, NULL, 'f'},
-			{"help", no_argument, NULL, 'h'},
-			{"remove", no_argument, NULL, 'r'},
+			{"force",        no_argument,       NULL, 'f'},
+			{"help",         no_argument,       NULL, 'h'},
+			{"remove",       no_argument,       NULL, 'r'},
+			{"root",         required_argument, NULL, 'R'},
+#ifdef WITH_SELINUX
+			{"selinux-user", no_argument,       NULL, 'Z'},
+#endif				/* WITH_SELINUX */
 			{NULL, 0, NULL, '\0'}
 		};
-		while ((c = getopt_long (argc, argv, "fhr",
+		while ((c = getopt_long (argc, argv,
+#ifdef WITH_SELINUX             
+		                         "fhrR:Z",
+#else				/* !WITH_SELINUX */
+		                         "fhrR:",
+#endif				/* !WITH_SELINUX */
 		                         long_options, NULL)) != -1) {
 			switch (c) {
 			case 'f':	/* force remove even if not owned by user */
 				fflg = true;
 				break;
+			case 'h':
+				usage (E_SUCCESS);
+				break;
 			case 'r':	/* remove home dir and mailbox */
 				rflg = true;
 				break;
+			case 'R': /* no-op, handled in process_root_flag () */
+				break;
+#ifdef WITH_SELINUX             
+			case 'Z':
+				if (is_selinux_enabled () > 0) {
+					Zflg = true;
+				} else {
+					fprintf (stderr,
+					         _("%s: -Z requires SELinux enabled kernel\n"),
+					         Prog);
+
+					exit (E_BAD_ARG);
+				}
+				break;
+#endif				/* WITH_SELINUX */
 			default:
-				usage ();
+				usage (E_USAGE);
 			}
 		}
 	}
 
 	if ((optind + 1) != argc) {
-		usage ();
+		usage (E_USAGE);
 	}
 
-	OPENLOG ("userdel");
-
 #ifdef ACCT_TOOLS_SETUID
 #ifdef USE_PAM
 	{
@@ -860,20 +949,23 @@ int main (int argc, char **argv)
 		retval = pam_acct_mgmt (pamh, 0);
 	}
 
-	if (NULL != pamh) {
-		(void) pam_end (pamh, retval);
-	}
 	if (PAM_SUCCESS != retval) {
-		fprintf (stderr, _("%s: PAM authentication failed\n"), Prog);
+		fprintf (stderr, _("%s: PAM: %s\n"),
+		         Prog, pam_strerror (pamh, retval));
+		SYSLOG((LOG_ERR, "%s", pam_strerror (pamh, retval)));
+		if (NULL != pamh) {
+			(void) pam_end (pamh, retval);
+		}
 		exit (E_PW_UPDATE);
 	}
+	(void) pam_end (pamh, retval);
 #endif				/* USE_PAM */
 #endif				/* ACCT_TOOLS_SETUID */
 
 	is_shadow_pwd = spw_file_present ();
 #ifdef SHADOWGRP
 	is_shadow_grp = sgr_file_present ();
-#endif
+#endif				/* SHADOWGRP */
 
 	/*
 	 * Start with a quick check to see if the user exists.
@@ -890,12 +982,18 @@ int main (int argc, char **argv)
 			              "deleting user not found",
 			              user_name, AUDIT_NO_ID,
 			              SHADOW_AUDIT_FAILURE);
-#endif
+#endif				/* WITH_AUDIT */
 			exit (E_NOTFOUND);
 		}
 		user_id = pwd->pw_uid;
+		user_gid = pwd->pw_gid;
 		user_home = xstrdup (pwd->pw_dir);
 	}
+#ifdef WITH_TCB
+	if (shadowtcb_set_user (user_name) == SHADOWTCB_FAILURE) {
+		exit (E_NOTFOUND);
+	}
+#endif				/* WITH_TCB */
 #ifdef	USE_NIS
 
 	/*
@@ -906,20 +1004,32 @@ int main (int argc, char **argv)
 		char *nis_master;
 
 		fprintf (stderr,
-			 _("%s: user %s is a NIS user\n"), Prog, user_name);
+		         _("%s: user %s is a NIS user\n"), Prog, user_name);
 		if (   !yp_get_default_domain (&nis_domain)
 		    && !yp_master (nis_domain, "passwd.byname", &nis_master)) {
 			fprintf (stderr,
-				 _("%s: %s is the NIS master\n"),
-				 Prog, nis_master);
+			         _("%s: %s is the NIS master\n"),
+			         Prog, nis_master);
 		}
 		exit (E_NOTFOUND);
 	}
-#endif
+#endif				/* USE_NIS */
 	/*
 	 * Check to make certain the user isn't logged in.
+	 * Note: This is a best effort basis. The user may log in between,
+	 * a cron job may be started on her behalf, etc.
 	 */
-	user_busy (user_name, user_id);
+	if (user_busy (user_name, user_id) != 0) {
+		if (!fflg) {
+#ifdef WITH_AUDIT
+			audit_logger (AUDIT_DEL_USER, Prog,
+			              "deleting user logged in",
+			              user_name, AUDIT_NO_ID,
+			              SHADOW_AUDIT_FAILURE);
+#endif				/* WITH_AUDIT */
+			exit (E_USER_BUSY);
+		}
+	}
 
 	/*
 	 * Do the hard stuff - open the files, create the user entries,
@@ -930,15 +1040,23 @@ int main (int argc, char **argv)
 	update_groups ();
 
 	if (rflg) {
-		remove_mailbox ();
+		errors += remove_mailbox ();
 	}
-	if (rflg && !fflg && (is_owner (user_id, user_home) == 0)) {
-		fprintf (stderr,
-			 _("%s: %s not owned by %s, not removing\n"),
-			 Prog, user_home, user_name);
-		rflg = 0;
-		errors++;
-		/* continue */
+	if (rflg) {
+		int home_owned = is_owner (user_id, user_home);
+		if (-1 == home_owned) {
+			fprintf (stderr,
+			         _("%s: %s home directory (%s) not found\n"),
+			         Prog, user_name, user_home);
+			rflg = 0;
+		} else if ((0 == home_owned) && !fflg) {
+			fprintf (stderr,
+			         _("%s: %s not owned by %s, not removing\n"),
+			         Prog, user_home, user_name);
+			rflg = 0;
+			errors++;
+			/* continue */
+		}
 	}
 
 #ifdef EXTRA_CHECK_HOME_DIR
@@ -959,9 +1077,8 @@ int main (int argc, char **argv)
 			}
 			if (path_prefix (user_home, pwd->pw_dir)) {
 				fprintf (stderr,
-					 _
-					 ("%s: not removing directory %s (would remove home of user %s)\n"),
-					 Prog, user_home, pwd->pw_name);
+				         _("%s: not removing directory %s (would remove home of user %s)\n"),
+				         Prog, user_home, pwd->pw_name);
 				rflg = false;
 				errors++;
 				/* continue */
@@ -970,13 +1087,13 @@ int main (int argc, char **argv)
 		}
 		endpwent ();
 	}
-#endif
+#endif				/* EXTRA_CHECK_HOME_DIR */
 
 	if (rflg) {
-		if (remove_tree (user_home) != 0) {
+		if (remove_tree (user_home, true) != 0) {
 			fprintf (stderr,
-				 _("%s: error removing directory %s\n"),
-				 Prog, user_home);
+			         _("%s: error removing directory %s\n"),
+			         Prog, user_home);
 			errors++;
 			/* continue */
 		}
@@ -988,7 +1105,7 @@ int main (int argc, char **argv)
 			              user_name, (unsigned int) user_id,
 			              SHADOW_AUDIT_SUCCESS);
 		}
-#endif
+#endif				/* WITH_AUDIT */
 	}
 #ifdef WITH_AUDIT
 	if (0 != errors) {
@@ -997,19 +1114,24 @@ int main (int argc, char **argv)
 		              user_name, AUDIT_NO_ID,
 		              SHADOW_AUDIT_FAILURE);
 	}
-#endif
+#endif				/* WITH_AUDIT */
 
 #ifdef WITH_SELINUX
-	if (is_selinux_enabled () > 0) {
-		const char *args[5];
-		args[0] = "/usr/sbin/semanage";
-		args[1] = "login";
-		args[2] = "-d";
-		args[3] = user_name;
-		args[4] = NULL;
-		safe_system (args[0], args, NULL, 1);
+	if (Zflg) {
+		if (del_seuser (user_name) != 0) {
+			fprintf (stderr,
+			         _("%s: warning: the user name %s to SELinux user mapping removal failed.\n"),
+			         Prog, user_name);
+#ifdef WITH_AUDIT
+			audit_logger (AUDIT_ADD_USER, Prog,
+			              "removing SELinux user mapping",
+			              user_name, (unsigned int) user_id,
+			              SHADOW_AUDIT_FAILURE);
+#endif				/* WITH_AUDIT */
+			fail_exit (E_SE_UPDATE);
+		}
 	}
-#endif
+#endif				/* WITH_SELINUX */
 
 	/*
 	 * Cancel any crontabs or at jobs. Have to do this before we remove
@@ -1018,6 +1140,10 @@ int main (int argc, char **argv)
 	user_cancel (user_name);
 	close_files ();
 
+#ifdef WITH_TCB
+	errors += remove_tcbdir (user_name, user_id);
+#endif				/* WITH_TCB */
+
 	nscd_flush_cache ("passwd");
 	nscd_flush_cache ("group");
 
